;;;
;;; Práctica de Introducción a la Inteligencia Artifical
;;; Curso: 3 ETSII-ULL
;;; Profesor: Patricio García
;;; Titulo: Sistema de Producción para el problema de 
;;;         Trasvase de agua entre cubos
;;; Fecha: Abril,1998
;;; Necesario: Librería "shell.lisp" para creación de Sistemas
;;;            de producción (LOAD "shell")
;;;            Librería "generic-search.lisp" para búsquedas genéricas
;;;            (LOAD "generic-search")
;;;
;;; Uso:
#|
;;;       Cargar primero:

(load "shell")
(load "generic-search")
(load "cubos")

;;;       Para búsqueda en anchura:

(ba)

;;;       Para búsqueda en profundidad:

(bp)

;;;       Ojo, se cicla por no tener tratamiento de estados visitados
|#

;;;
;;; Modificaciones:
;;;    Mayo 2004: 
;;;       Cambios debidos a modificación de q_search para 
;;;       llamar a generic_search con una lista con el estado
;;;       inicial.
;;;       Cambios para que IMP-EST devuelva el estado que imprime
;;;       Cambios debidos a modificación de shell.lsp para llamar 
;;;       a shell como paquete
;;;    Mayo 2006:
;;;       Cambio de texto mayúsculas/minúsculas
;;;    Mayo 2007:
;;;       Cambio nombre Q_SEARCH a GENERIC-SEARCH, "cubos.lsp" a 
;;;       "cubos.lisp"
;;;
;;; Pendiente por hacer:
;;;       Incluir devolución de camino
;;;       Tratamiento de estados visitados
;;;

; Usa librerías para reglas y para busquedas



; Constante versión

(defconstant version-cubos "may-2009"
  "Versión del programas cubos.")


; Estructura para estados
; -----------------------

(defstruct estado
  "Estructura para estados del problema de los cubos."
  (cubo1 0)     ; Cantidad de litros de agua del cubo1
  (cubo2 0))    ; Cantidad de litros de agua del cubo2


; Variables globales 
; ------------------

;Capacidades de los cubos

(defvar capacidad1 8 
  "Capacidad por defecto del cubo1.")

(defvar capacidad2 6 
  "Capacidad por defecto del cubo2.")


;Litros finales deseados en uno de los cubos

(defvar litrosfinales 4
  "Litros finales a obtener en el estado objetivo.")


; Definición de lista de estados iniciales

(defvar estados-iniciales (list (make-estado))
  "Lista de estados iniciales.")


;Definicion de condición de finalización de búsqueda
; --------------------------------------------------

(defun estados-objetivo-p (estado)
  "Chequea si ha de detenerse la búsqueda en el problema de los cubos."
  (if estado
      (or (= (estado-cubo1 estado) litrosfinales)
	  (= (estado-cubo2 estado) litrosfinales))
      t))


; Definición de funciones para las reglas
; ---------------------------------------

;Regla1: Llena cubo 1

(defun p1 (est)
  "Precondición regal 1."
  (/= (estado-cubo1 est) capacidad1))

(defun a1 (est)
  "Acción regla 1."
  (make-estado :cubo2 (estado-cubo2 est)
	       :cubo1 capacidad1))


;Regla2: Llena cubo 2

(defun p2  (est)
  "Precondición regal 2."
  (/= (estado-cubo2 est) capacidad2))

(defun a2 (est)
  "Acción regla 2."
  (make-estado :cubo1 (estado-cubo1 est)
	       :cubo2 capacidad2))

;Regla3: Vacia cubo 1
(defun p3  (est)
  "Precondición regal 3."
  (/= (estado-cubo1 est) 0))

(defun a3 (est)
  "Acción regla 3."
  (make-estado :cubo2 (estado-cubo2 est)
		 :cubo1 0))


;Regla4: Vacia cubo 2
(defun p4  (est)
  "Precondición regal 4."
  (/= (estado-cubo2 est) 0))

(defun a4 (est)
  "Acción regla 4."
  (make-estado :cubo1 (estado-cubo1 est)
	       :cubo2 0))


;Regla5: Trasvasa del 1 al 2
(defun p5  (est)
  "Precondición regal 5."
  (and (/= (estado-cubo1 est) 0)
       (/= (estado-cubo2 est) capacidad2)))

(defun a5 (est)
  "Acción regla 5."
  (let ((cant (min (estado-cubo1 est) 
		   (- capacidad2 
		      (estado-cubo2 est)))))
    (make-estado :cubo1 (- (estado-cubo1 est) cant)
		 :cubo2 (+ (estado-cubo2 est) cant))))


;Regla6: Trasvasa del 2 al 1
(defun p6  (est)
  "Precondición regal 6."
  (and (/= (estado-cubo1 est) capacidad1)
       (/= (estado-cubo2 est) 0)))

(defun a6 (est)
  "Acción regla 6."
  (let ((cant (min (estado-cubo2 est) 
		   (- capacidad1 
		      (estado-cubo1 est)))))
    (make-estado :cubo1 (+ (estado-cubo1 est) cant)
		 :cubo2 (- (estado-cubo2 est) cant))))


; Inserción las reglas en el sistema de producción
; ------------------------------------------------
(shell:usa-sistema 'cubos)

(shell:regla #'p1 #'a1)
(shell:regla #'p2 #'a2)
(shell:regla #'p3 #'a3)
(shell:regla #'p4 #'a4)
(shell:regla #'p5 #'a5)
(shell:regla #'p6 #'a6)


; Función para imprimir un estado
; -------------------------------

(defun imp-est (est)
  "Imprime un estado del problema de los cubos."
  (if est
      (progn
	(format t "[~d ~d] " (estado-cubo1 est) (estado-cubo2 est))
	est)
    nil))



;Llamada a búsqueda en profundidad
; --------------------------------

(defun bp ()
  "Búsqueda en profundidad del problema de los cubos."
  (imp-est 
   (generic-search:generic-search estados-iniciales
		   #'estados-objetivo-p
		   #'shell:hijos
		   :display-fn #'imp-est
		   :merge-fn #'(lambda (new-states)
				 (nconc new-states generic-search:*search-queue*))
		   :dequeue-fn #'(lambda () (pop generic-search:*search-queue*)))))


;Llamada a búsqueda en anchura
; ----------------------------

(defun ba ()
  "Búsqueda en anchura del problema de los cubos."
  (imp-est 
   (generic-search:generic-search estados-iniciales
		   #'estados-objetivo-p
		   #'shell:hijos
		   :display-fn #'imp-est
		   :merge-fn #'(lambda (new-states)
				 (nconc generic-search:*search-queue* new-states))
		   :dequeue-fn #'(lambda () (pop generic-search:*search-queue*)))))

